from ltda.ltda_common import *
from ltda.ltda_tools import Downloader
import ltda.ltda_config
#sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../CDBAPI/Python/')))
from DatabaseApi import *
from commvault_ares.archiver import Archive
import logging


logger = logging.getLogger('ltda_archiver')
# logger.setLevel(logging.INFO)
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

if len(logger.handlers) == 0:
    logger.addHandler(handler)
else:
    # avoid adding multiple handlers which would cause one message to be printed multiple times
    logger.handlers[0] = handler

#===============================================================================
# Classes definitions
class LTDA_Archiver(AbstractPoller):
    ''' @brief    Archiver class, containing polling loop
    '''
    def __init__(self):
        ''' @brief    Class constructor
        '''
        AbstractPoller.__init__(self, ltda.ltda_config.POLL_PERIOD_ARCHIVER)

    def startProductArchive(self, product):

        product_type = str(product.getProductType())
        if product_type in ltda.ltda_config.PROD_TYPE_SKIP_LST:
            #skip this type of product
            logger.debug("Skipping product %s." % product.getProductReference())
            return
        logger.debug("Archiving product %s." % product.getProductReference())

        #we should archive the dmz location because this is the location where the metadata is located too.
        # Only for TFF and L1A there is not dmz location. For these products the dmz location is empty!
        if product_type in ['PROBAV_L1A', 'PROBAV_TFF']:
            archive_source_location = str(product.getProductLocation())
        else:
            archive_source_location = str(product.getProductLocationDmz())


        if archive_source_location == "":
            logger.error("Product location is empty, we are unable to Archive this.")
            archiveHistory = ArchiveHistory(ArchiveStatus.ArchiveError,
                                            EasyGregorianCalendar(Date()),
                                            "0",
                                            'empty')
            self.cdbApi.setArchiveStatus(product.getProductReference(), archiveHistory)
            raise Exception("Product location is empty, we are unable to Archive this.")

        subclient, subdir = getSubclient(product_type)

        #parse product start time
        product_start_time = product.getProductStartTime().getTime()
        df = DateFormatter("yyyy-MM-dd")
        dbDate = str(df.format(product_start_time)).split('-')
        year = dbDate[0]
        month = dbDate[1]
        day = dbDate[2]
        archive_subdir = os.path.join(subdir, year, month, day) # dir created in the subclient folder

        #initiate archive
        arc = Archive(archive_source_location, subclient, archive_subdir)
        status = arc.check_source_location()    #status=0: OK, so proceed

        if status == 0:
            #update DB, set status archiving
            self.cdbApi.setArchiveStatus(product.getProductReference(),
                                        ArchiveHistory(ArchiveStatus.Archiving,
                                                       EasyGregorianCalendar(Date()),
                                                       "0",
                                                       os.path.join(arc.cache_location, arc.source_fname)))

            # Start downloading the data
            dl = Downloader(archive_source_location, arc.cache_location)
            status, cache_location = dl.caching_source()

        if status != 0:
            logger.error("Unable to cache")
            self.cdbApi.setArchiveStatus(product.getProductReference(),
                                         ArchiveHistory(ArchiveStatus.ArchiveError,
                                                        EasyGregorianCalendar(Date()),
                                                        "0",
                                                        ''))
            return

        status, archive_location = dl.move_to_final(arc.archive_location, arc.source_fname)
        if status != 0:
            logger.error("Unable to move to final location")
            cdb_status = ArchiveStatus.ArchiveError
        else:
            cdb_status = ArchiveStatus.Archived

        logger.debug("Archiving operation is over, Updating with status %s " % cdb_status)
        self.cdbApi.setArchiveStatus(product.getProductReference(),
                                     ArchiveHistory(cdb_status,
                                                    EasyGregorianCalendar(Date()),
                                                    "0",
                                                    archive_location))

    def _handle_calibration_product(self, calibration):
        logger.debug("Processing calibration data %s." % calibration.getReference())
        calibrationProductType = str(calibration.getCalibrationProductType())
        archive_source_location = str(calibration.getLocation())
        subclient, subdir = getSubclient(calibrationProductType)

        # parse product start time
        product_start_time = calibration.getStartTime().getTime()
        df = DateFormatter("yyyy-MM-dd")
        dbDate = str(df.format(product_start_time)).split('-')
        year = dbDate[0]
        month = dbDate[1]
        day = dbDate[2]
        archive_subdir = os.path.join(subdir, year, month, day)  # dir created in the subclient folder

        if archive_source_location == "":
            raise Exception("There is no archive_source_location for calibration product")

        arc = Archive(archive_source_location, subclient, archive_subdir)
        status, archive_location = arc.do_archive()
        if status != 0:
            logger.error("Unable to archive ")
            cdb_status = ArchiveStatus.ArchiveError
        else:
            cdb_status = ArchiveStatus.Archived

        logger.debug("Archiving operation is over, Updating with status %s " % cdb_status)
        calArchiveHistory = CalibrationDataArchiveHistory(cdb_status,
                                                          EasyGregorianCalendar(Date()),
                                                          "0",
                                                          archive_location)
        self.cdbApi.setCalibrationDataArchiveStatus(calibration.getProductReference(), calArchiveHistory)

    def list_and_start_archive(self):
        ''' @brief    Polling CDB for new data to archive
        '''
        # Poll CDB for products
        logger.debug("Polling the CDB for products to be archived...")
        products = self.cdbApi.getProductsByStatus(None,
                                                   None,
                                                   ProductStatus.Online,
                                                   None,
                                                   None,
                                                   ArchiveStatus.ToBeArchived,
                                                   None)
        for product in products:
            self.startProductArchive(product)

        return

    def poll(self):
        while True:
            self.list_and_start_archive()
            print("Sleep for some time....")
            time.sleep(self.period)

#===============================================================================
# Actual program launch
def main():
    ''' @brief    LTDA Archiver program entry point
    '''
    try:
        logger.debug("Launching Archiver...")
    except:
        pass

    arc = LTDA_Archiver()
    arc.poll()
    return arc


#===============================================================================
# Global exceptions logging and management
if __name__ == '__main__':
    main()
