import os
import time

import ltda.ltda_config
from ltda.ltda_tools import CachedData
from DatabaseApi import *
from ltda.ltda_producttypes import project_dict

import logging
logger = logging.getLogger(__name__)

#===============================================================================
# Classes definitions
class AbstractPoller(object):
    ''' @brief    Abstract class for polling CDB , meant to be subclassed
    '''
    def __init__(self, period):
        ''' @brief    Class constructor
            @param    period: time period in seconds between each poll
        '''
        object.__init__(self)
        self.period = period
        self.activeRestoringProducts = {}

        # Connection with the CDB
        try:
            if 'HIBERNATE_CFG_PATH' in os.environ:
                cfg_path = os.getenv('HIBERNATE_CFG_PATH')
            else:
                cfg_path = os.path.join(os.getcwd(), ltda.ltda_config.HIBERNATE_CFG_PATH)

            print(cfg_path)
            self.cdbApi = DataFacadeFactory.createDataFacade(cfg_path);
            logger.debug("Connected to CDB")
        except:
            logger.error("Unable to connect to the CDB!")
            raise

    def startPollLoop(self):
        ''' @brief    Launch polling loop
        '''
        while True:
            self.__doPoll()
            logger.debug("Sleep some time...")
            time.sleep(self.period)

    def __doPoll(self):
        ''' @brief    Execute poll actions
        '''
        logger.debug( "Starting Polling Action...")

        self.list_active_restores()
        self.__checkCVjob()
        self._pollCdb()




class AbstractTicket(object):
    '''@brief A ticket issued by the CAF, abstract class intended to be subclassed
    '''
    # KVH 20112013 Handle group tickets
    def __init__(self, jobid, cdbApi, final_restore_location, commvault_restore_location=None):
        '''@brief    Class constructor
           @param    ticketNumber: the number of the ticket issued by the CAF
           @param    ltdaPending: True if processing needed, False to discard
           @param    final_restore_location: Temporary restore location for CAF
        '''
        object.__init__(self)
        self.jobid = jobid
        self.commvault_restore_location = str(commvault_restore_location)
        self.cdbApi = cdbApi
        self.final_restore_location = final_restore_location
        self.ltdaPending = True
        self.creationTime = time.time()
        self.attemptNumber = 1
        self.creationTime = time.time()
        self.status = None


class ProductTicket(AbstractTicket):
    '''@brief A jobid issued by Commvault for a product
    '''

    def __init__(self, jobid, product, cdbApi, final_restore_location = None, commvault_restore_location = None):
        '''@brief    Class constructor
           @param    ticketNumber: the number of the ticket issued by the CAF
           @param    ltdaPending: True if processing needed, False to discard
           @param    objUrl: Temporary restore location for CAF
        '''
        AbstractTicket.__init__(self, jobid, cdbApi, final_restore_location, commvault_restore_location)
        self.product = product

    def resolve(self, status=None):
        ''' Handle cached data(FTP to final location) and update the archivehistory and archivestatus'''

        self.status = status
        #dateformatter = DateFormatter("yyyyMMdd")
        logger.debug("status : " + str(self.status))

        #status=0 : completed; status=1 :running; status=-1 :failed
        if self.status == 0:
            # Success: close ticket
            if not self.product.getArchiveStatus() == ArchiveStatus.Restoring:
                raise Exception("Resolving is in an unknown state : %s" % self.product.getArchiveStatus())


            final_restore_location = str(self.final_restore_location)

            #move data from cachedir
            logger.debug("Moving from cached dir...")
            cd = CachedData(final_restore_location)
            restore_location = cd.move_to_final_location(self.commvault_restore_location)

            # Update archive status
            logger.debug("Restoring operation succeeded. Updating status...")
            archiveHistory = ArchiveHistory(ArchiveStatus.Restored,
                                            EasyGregorianCalendar(Date()),
                                            self.jobid,
                                            restore_location)

            self.cdbApi.setArchiveStatus(self.product.getProductReference(), archiveHistory)

            # Update object reference
            self.product = self.cdbApi.getProductByReference(self.product.getProductReference())
            logger.info("Ticket %s, reference %s successfully closed." % (self.jobid, self.getReference()))
            self.ltdaPending = False
        else:
            # Update archive status

            archiveHistory = ArchiveHistory(ArchiveStatus.ArchiveError,
                                            EasyGregorianCalendar(Date()),
                                            self.jobid,
                                            None)

            self.cdbApi.setArchiveStatus(self.product.getProductReference(), archiveHistory)

            # Update object reference
            self.product = self.cdbApi.getProductByReference(self.product.getProductReference())
            logger.info("Ticket %s, reference %s closed with error." % (self.jobid, self.getReference()))
            self.ltdaPending = False

    def getReference(self):
        '''@brief    Retrieve the product or calibration reference
        '''
        return self.product.getProductReference()

    @staticmethod
    def getLastArchiveFromHistory(product, cdbApi):
        ''' @brief    Get last archive entry
            @return   Last entry for this current object from its archive history if exists, None otherwise
        '''
        productReference = product.getProductReference()
        archiveHistory = cdbApi.getArchiveHistory(productReference)

        lastArchive = None
        lastDate = None
        for archiveStatus in archiveHistory:
            currentDate = archiveStatus.getStatusActivation()
            if lastDate is None or currentDate.after(lastDate):
                lastDate = currentDate
                lastArchive = archiveStatus
        return lastArchive

    @staticmethod
    def getLastArchivedStatusFromHistory(product, cdbApi):
        ''' @brief    Get last archive entry
            @return   Last entry for this current object from its archive history if exists, None otherwise
        '''
        productReference = product.getProductReference()
        archiveHistory = cdbApi.getArchiveHistory(productReference)

        lastArchive = None
        lastDate = None
        for archiveStatus in archiveHistory:
            currentDate = archiveStatus.getStatusActivation()
            if archiveStatus.getArchiveStatus() == ArchiveStatus.Archived:
                if lastDate is None or currentDate.after(lastDate):
                    lastDate = currentDate
                    lastArchive = archiveStatus
        return lastArchive

    @staticmethod
    def getLastRestoringStatusFromHistory(product, cdbApi):
        ''' @brief    Get last archive entry
            @return   Last entry for this current object from its archive history if exists, None otherwise
        '''
        productReference = product.getProductReference()
        archiveHistory = cdbApi.getArchiveHistory(productReference)

        lastArchive = None
        lastDate = None
        for archiveStatus in archiveHistory:
            currentDate = archiveStatus.getStatusActivation()
            if archiveStatus.getArchiveStatus() == ArchiveStatus.Restoring:
                if lastDate is None or currentDate.after(lastDate):
                    lastDate = currentDate
                    lastArchive = archiveStatus
        return lastArchive


# class CalibrationTicket(AbstractTicket):
#     '''@brief A ticket issued by the CAF for a calibration
#     '''
#     def __init__(self, ticketNumber, calibration, cdbApi, objUrl = None):
#         '''@brief    Class constructor
#            @param    ticketNumber: the number of the ticket issued by the CAF (GetObj ticket number)
#            @param    ltdaPending: True if processing needed, False to discard
#            @param    objUrl: Temporary restore location for CAF
#         '''
#         AbstractTicket.__init__(self, ticketNumber, cdbApi, objUrl)
#         self.calibration = calibration
#
#     def resolve(self):
#         '''@brief    Closes the operation and put in history
#         '''
#         if self.status is None:
#             self.checkStatus()
#
#         if self.status == "0":
#             # Success: close ticket
#             if self.calibration.getArchiveStatus() == ArchiveStatus.Archiving:
#                 logger.debug( "Archiving operation succeeded. Updating status...")
#                 calArchiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.Archived,
#                                                                   EasyGregorianCalendar(Date()),
#                                                                   self.ticketNumber,
#                                                                   self.calibration.getLocation())
#                 self.cdbApi.setCalibrationDataArchiveStatus(self.calibration.getReference(), calArchiveHistory)
#                 self.calibration = self.cdbApi.getCalibrationDataByReference(self.calibration.getReference())
#
#                 # Delete the metadata file if needed
#                 if ltda_config.METADATA_DELETE:
#                     logger.debug( "Deleting Metadata file...")
#                     try:
#                         # JME : 19042011 : replace # in files by "_"
#                         ref = self.calibration.getReference().replace("#","_")
#                         os.remove(ltda_config.METADATA_DIR + os.sep + ltda_config.METADATA_CALIB_PATTERN % ref)
#                     except:
#                         logger.warning('Unable to delete log file for calibration reference "%s"!' % ref)
#                 logger.info("Ticket %s, reference %s successfully closed." % (self.ticketNumber, self.getReference()))
#                 self.ltdaPending = False
#             elif self.calibration.getArchiveStatus() == ArchiveStatus.Restoring:
#                 logger.debug("Restoring operation succeeded. Updating status...")
#
#                 # TVR: JIRA PRV-371/372/373 (moving file to correct location)
#
#                 filename = self._moveObjToFinalLocation()
#                 location = os.path.dirname(self.objUrl) + "/" + filename;  # Move one directory up and append filename
#
#                 # Update archive status
#                 calArchiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.Restored,
#                                                                   EasyGregorianCalendar(Date()),
#                                                                   self.ticketNumber,
#                                                                   location)
#                 self.cdbApi.setCalibrationDataArchiveStatus(self.calibration.getReference(), calArchiveHistory)
#                 # Update object reference
#                 self.calibration = self.cdbApi.getCalibrationDataByReference(self.calibration.getReference())
#                 logger.info("Ticket %s, reference %s successfully closed." % (self.ticketNumber, self.getReference()))
#                 self.ltdaPending = False
#         elif self.status == "1" or self.status == "7" or self.status == "8":
#             # nothing to do, just for next poll except
#             pass
#         elif self.status == "-1" and self.attemptNumber <= ltda_config.MAX_CAF_REPOST:
#             # Failure: Discard ticket and set status to initial one.
#             if self.calibration.getArchiveStatus() == ArchiveStatus.Archiving:
#                 logger.warning("Archiving operation of ticket %s, reference %s failed for unknown reason. Retrying..." % (self.ticketNumber, self.getReference()))
#                 calArchiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.ToBeArchived,
#                                                                EasyGregorianCalendar(Date()),
#                                                                self.ticketNumber,
#                                                                self.calibration.getLocation())
#                 self.cdbApi.setCalibrationDataArchiveStatus(self.calibration.getReference(), calArchiveHistory)
#                 self.calibration = self.cdbApi.getCalibrationDataByReference(self.calibration.getReference())
#                 self.ltdaPending = False
#                 self.attemptNumber += 1
#             elif self.calibration.getArchiveStatus() == ArchiveStatus.Restoring:
#                 logger.warning("Restoring operation of ticket %s, reference %s failed for unknown reason. Retrying..." % (self.ticketNumber, self.getReference()))
#                 calArchiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.ToBeRestored,
#                                                                   EasyGregorianCalendar(Date()),
#                                                                   self.ticketNumber,
#                                                                   self.calibration.getLocation())
#                 self.cdbApi.setCalibrationDataArchiveStatus(self.calibration.getReference(), calArchiveHistory)
#                 self.calibration = self.cdbApi.getCalibrationDataByReference(self.calibration.getReference())
#                 self.ltdaPending = False
#                 self.attemptNumber += 1
#
#         # JME 20042011 Unknown ticket management
#         elif self.status == "UNKNOWN_TICKET":
#             if self.product.getArchiveStatus() == ArchiveStatus.Archiving:
#                 logger.error("Unable to archive calibration data '%s' (reason: unknown archive ticket %s), waivering unknown jobid." % (self.getReference(), self.ticketNumber))
#
#                 reference = self.product.getProductReference()
#                 currentDate = EasyGregorianCalendar(Date())
#                 archiveHistory = ArchiveHistory(ArchiveStatus.ArchiveError, currentDate, "", None)
#                 self.cdbApi.setArchiveStatus(reference, archiveHistory)
#
#                 # Delete the metadata file if needed
#                 if ltda_config.METADATA_DELETE:
#                     logger.debug("Deleting Metadata file...")
#                     try:
#                         # JME : 19042011 : replace # in files by "_"
#                         ref = self.product.getProductReference().replace("#","_")
#                         os.remove(ltda_config.METADATA_DIR + os.sep + ltda_config.METADATA_PROD_PATTERN % ref)
#                     except:
#                         logger.warning('Unable to delete log file for product reference "%s"!' % ref)
#                 self.ltdaPending = False
#
#             elif self.product.getArchiveStatus() == ArchiveStatus.Restoring:
#                 logger.error("Unable to restore calibration data '%s' (reason: unknown restore ticket %s), waivering unknown jobid." % (self.getReference(), self.ticketNumber))
#
#                 reference = self.product.getProductReference()
#                 currentDate = EasyGregorianCalendar(Date())
#                 archiveHistory = ArchiveHistory(ArchiveStatus.RestoreError, currentDate, "", None)
#                 self.cdbApi.setArchiveStatus(reference, archiveHistory)
#                 self.ltdaPending = False
#
#         else:
#             if self.status == "-2":
#                 logger.warning("Ticket %s, reference %s is bad! Aborting." % (self.ticketNumber, self.getReference()))
#             elif self.attemptNumber > ltda_config.MAX_CAF_REPOST:
#                 logger.warning("Ticket %s, reference %s failed %d times. Aborting." % (self.ticketNumber, self.getReference(), self.attemptNumber))
#             # Setting the status to an error one
#             if self.calibration.getArchiveStatus() == ArchiveStatus.Archiving:
#                 archiveStatus = ArchiveStatus.ArchiveError
#             elif self.calibration.getArchiveStatus() == ArchiveStatus.Restoring:
#                 archiveStatus = ArchiveStatus.RestoreError
#             else:
#                 logger.error("Archive Status not valid for calibration %s." % self.calibration.getReference())
#                 return
#             calArchiveHistory = CalibrationDataArchiveHistory(archiveStatus,
#                                                               EasyGregorianCalendar(Date()),
#                                                               self.ticketNumber,
#                                                               self.calibration.getLocation())
#             self.cdbApi.setCalibrationDataArchiveStatus(self.calibration.getReference(), calArchiveHistory)
#             self.calibration = self.cdbApi.getCalibrationDataByReference(self.calibration.getReference())
#             self.ltdaPending = False
#
#     def getReference(self):
#         '''@brief    Retrieve the product or calibration reference
#         '''
#         return self.calibration.getReference()
#
#     @staticmethod
#     def getLastArchiveFromHistory(calibration, cdbApi):
#         ''' @brief    Get last archive entry
#             @return   Last entry for this current object from its archive history if exists, None otherwise
#         '''
#
#         calibrationReference = calibration.getReference()
#         archiveHistory = cdbApi.getCalibrationDataArchiveHistory(calibrationReference)
#
#         lastArchive = None
#         lastDate = None
#         for archiveStatus in archiveHistory:
#             currentDate = archiveStatus.getStatusActivation()
#             if lastDate is None or currentDate.after(lastDate):
#                 lastDate = currentDate
#                 lastArchive = archiveStatus
#         return lastArchive
#
#     @staticmethod
#     def getLastArchivedStatusFromHistory(calibration, cdbApi):
#         ''' @brief    Get last archive entry
#             @return   Last entry for this current object from its archive history if exists, None otherwise
#         '''
#         calibrationReference = calibration.getReference()
#         archiveHistory = cdbApi.getCalibrationDataArchiveHistory(calibrationReference)
#
#         lastArchive = None
#         lastDate = None
#         for archiveStatus in archiveHistory:
#             currentDate = archiveStatus.getStatusActivation()
#             if archiveStatus.getArchiveStatus() == ArchiveStatus.Archived:
#                 if lastDate is None or currentDate.after(lastDate):
#                     lastDate = currentDate
#                     lastArchive = archiveStatus
#         return lastArchive


def getSubclient( productType):
    ''' @brief    Get the commvault subclient for a given product type
        @param    productType: the product type used to find the project
        @return   the subclient name as a string
    '''

    productType=str(productType)



    # subclient_dict = {
    #     'meteo' : ['METEO_ECMWF','METEO_METEOSERVICES'],
    #     'cal' : ['POLARMOTION', 'PROBAV_L1A_CALIBRATION', 'PROBAV_L1B_CALIBRATION', 'PROBAV_L1B_OVERLAPREGION' ,
    #              'PROBAV_L1B_INTERSECTION', 'PROBAV_L1C_INTERSECTION',
    #              'GEOMETRIC_CALIBRATION_REPORT','RADIOMETRIC_CALIBRATION_REPORT','ICP_GEOMETRIC_CENTRE',
    #              'ICP_GEOMETRIC_LEFT','ICP_GEOMETRIC_RIGHT','ICP_RADIOMETRIC_CENTRE','ICP_RADIOMETRIC_LEFT',
    #              'ICP_RADIOMETRIC_RIGHT'
    #              ],
    #     'tff' : ['PROBAV_TRANSFERFRAMES'],
    #     'l1a' : ['PROBAV_L1A', 'PROBAV_L1A_ANTAR'],
    #     'l1c' : ['PROBAV_L1C','PROBAV_L1C_ANTAR'],
    #     'l1b' : ['PROBAV_L1B'],
    #     'l2x' : ['PROBAV_L2A_1KM','PROBAV_L2A_333M', 'PROBAV_L2A_100M','PROBAV_L2B_1KM',
    #             'PROBAV_L2B_333M', 'PROBAV_L2B_100M', 'PROBAV_L2A_333M_ANTAR', 'PROBAV_L2A_1KM_ANTAR',
    #              'PROBAV_L2A_100M_ANTAR'],
    #     'l3_s10_1km'     : ['PROBAV_L3_S10_TOC_1KM', 'PROBAV_L3_S10_TOC_NDVI_1KM'],
    #     'l3_s10_333m'    : ['PROBAV_L3_S10_TOC_333M', 'PROBAV_L3_S10_TOC_NDVI_333M'],
    #     'l3_s1_toa_100m' : ['PROBAV_L3_S1_TOA_100M'],
    #     'l3_s1_toa_1km'  : ['PROBAV_L3_S1_TOA_1KM'],
    #     'l3_s1_toa_333m' : ['PROBAV_L3_S1_TOA_333M'],
    #     'l3_s1_toc_100m' : ['PROBAV_L3_S1_TOC_100M'],
    #     'l3_s1_toc_1km'  : ['PROBAV_L3_S1_TOC_1KM'],
    #     'l3_s1_toc_333m' : ['PROBAV_L3_S1_TOC_333M'],
    #     'l3_sxx_100m'    : ['PROBAV_L3_S5_TOA_100M', 'PROBAV_L3_S5_TOC_100M', 'PROBAV_L3_S30_TOC_100M',
    #                         'PROBAV_L3_S1_TOC_NDVI_100M', 'PROBAV_L3_S5_TOC_NDVI_100M' ]}
    try:
        subclient = project_dict[productType]['subclient']
        subdir = project_dict[productType]['subdir']
    except KeyError:
        raise Exception("KEYERROR : %s" % productType)


    if subclient is None:
        raise Exception("Product type %s not found in subclient_dict" % productType)

    return subclient, subdir