# Module imports
import sys
import traceback

from ltda.ltda_common import *
from ltda.ltda_tools import *
import ltda.ltda_config
from ltda.ltda_common import AbstractPoller

from commvault_ares.restorer import Restorer
from commvault_ares.restorer import getCVStatus

#sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../../CDBAPI/Python/')))
from DatabaseApi import *

import logging

logger = logging.getLogger('ltda_restorer')
logger.setLevel(logging.DEBUG)

handler = logging.StreamHandler()
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)

if len(logger.handlers) == 0:
    logger.addHandler(handler)
else:
    # avoid adding multiple handlers which would cause one message to be printed multiple times
    logger.handlers[0] = handler


class LTDA_Restorer(AbstractPoller):
    ''' @brief    Restorer class, containing polling loop
    '''
    def __init__(self):
        ''' @brief    Class constructor
        '''
        AbstractPoller.__init__(self, ltda.ltda_config.POLL_PERIOD_RESTORER)

    def list_active_restores(self):
        ''' @brief    Check for unreferenced "restoring" products (e.g. when restarting the LTDA)
         and add them to the active list"
        '''
        logger.debug( "Polling the CDB for restoring products...")
        products = self.cdbApi.getProductsByStatus(None,
                                                   None,
                                                   None,
                                                   None,
                                                   None,
                                                   ArchiveStatus.Restoring,
                                                   None)
        for product in products:
            if not product.getProductReference() in self.activeRestoringProducts.keys():
                # Ticket number lost
                lastArchive = ProductTicket.getLastArchiveFromHistory(product, self.cdbApi)

                if lastArchive.getArchiveStatus() == ArchiveStatus.Restoring:
                    # put the ticket back in the activeticketslist
                    #getting all restoring information stored in DB
                    archive_location = ProductTicket.getLastArchivedStatusFromHistory(product, self.cdbApi).getLocation()
                    commvault_temp_restore_location = ProductTicket.getLastRestoringStatusFromHistory(product, self.cdbApi).getLocation()
                    commvaultJobNr = lastArchive.getTicketNumber()
                    final_restore_location = product.getProductLocation()


                    self.activeRestoringProducts[product.getProductReference()] = \
                        ProductTicket(commvaultJobNr,
                                      product,
                                      self.cdbApi,
                                      final_restore_location,
                                      commvault_restore_location = commvault_temp_restore_location)
                    logger.debug( "Ticket number %s was recreated..." % commvaultJobNr)
                # else:
                #     # Ticket definitively lost, restart restore
                #     archiveHistory = ArchiveHistory(ArchiveStatus.ToBeRestored,
                #                                     EasyGregorianCalendar(Date()),
                #                                     lastArchive.getTicketNumber(),
                #                                     product.getProductLocation())
                #     self.cdbApi.setArchiveStatus(product.getProductReference(), archiveHistory)
                #     logger.warning( "Unable to restore product '%s' (reason: ticket lost), setting it to archive status 'ToBeRestored' again." % product.getProductReference())


        return


    def _checkActiveRestores(self):
        '''  Checking current commvault job status and resolve them if the job is finished.
        '''

        logger.debug( "Checking current Commvault job status...")
        rest = None

        #loop over all active tickets and check its status
        for (ref, job) in self.activeRestoringProducts.items():
            jobid = str(job.jobid)
            logger.debug("jobIDToCheck : %s" % jobid)
            status, job_status, job_err_msg, rest = getCVStatus(jobid, rest)

            if status == 1: #ongoing
                #ticket will be checked later on again
                logger.debug("Commvault Job %s status: %s"  % ( jobid, job_status))
                waitTime = time.time() - job.creationTime
                if waitTime > ltda.ltda_config.COMMVAULT_TIMEOUT:
                    logger.warning( 'Commvault job %s, reference %s with status %s is running since %f seconds.' % (
                        jobid.ticketNumber,
                        ref,
                        job_status,
                        waitTime))
            elif status == -1: #Done with ERROR
                logger.warning( "Ticket %s, reference %s, status: %s" % (
                    jobid,
                    ref,
                    job_status
                    ))
                job.resolve(status)
            elif status == 0: #Done with success
                logger.debug("Ticket %s, reference %s, status: %s" % (
                    jobid,
                    ref,
                    job_status))
                job.resolve(status)
            else:
                logger.error("invalid status: %s" % (status))
                raise Exception(
                    "Commvault job Details for job '%s' is not as expected. Does the job exists?" % self.jobid)


        # Discard all tickets not pending (resolved or rejected)
        ticketsToDiscard = [ ref for (ref,ticket) in self.activeRestoringProducts.items() if not job.ltdaPending ]
        for ref in ticketsToDiscard:
            del self.activeRestoringProducts[ref]

    def _pollCdb(self):
        pass

    def startProductRestore(self):
        ''' @brief    Polling CDB for new data to restore
        '''
        logger.debug( "Polling the CDB for data to be restored...")
        products = self.cdbApi.getProductsByStatus(None,
                                                   None,
                                                   None,
                                                   None,
                                                   None,
                                                   ArchiveStatus.ToBeRestored,
                                                   None)
        productsToGet = {}


        for product in products:
            reference = product.getProductReference()
            product_type = str(product.getProductType())

            # tff L1A => productlocation, voor de rest productlocationdmz.
            #if product_type in ['PROBAV_L1A', 'PROBAV_TFF']:
            finalRestoreLocation = str(product.getProductLocation())
            if finalRestoreLocation == "" or finalRestoreLocation is None:
                 finalRestoreLocation = ltda.ltda_config.DEFAULT_PROD_REST_FTP
            #else:
            #    finalRestoreLocation = str(product.getProductLocationDmz())

            #use productlocation by default, if not exist, use productlocationdmz
            # finalRestoreLocation = str(product.getProductLocation())


            subclient, subdir = getSubclient(product_type)
            currentDate = EasyGregorianCalendar(Date())

            if ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS != 0:
                logger.debug( "Queue size limit: %d Tickets in queue: %d" % (
                    ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS, len(self.activeRestoringProducts)))
                if len(self.activeRestoringProducts) >= ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS:
                    break
            logger.debug( "Processing product %s." % product.getProductReference())

            #check if all locations are present in the database
            # get the archive_location from the last archived status from the archivehistory tabel
            lastArchive = ProductTicket.getLastArchivedStatusFromHistory(product, self.cdbApi)
            if lastArchive is None:
                logger.warning( "Unable to restore product '%s' (reason: never archived), setting it to archive status 'NotArchived' again." % reference)
                archiveHistory = ArchiveHistory(ArchiveStatus.NotArchived,
                                                currentDate,
                                                "",
                                                None)
                self.cdbApi.setArchiveStatus(product.getProductReference(), archiveHistory)
                continue
            archive_location = str(lastArchive.getLocation())
            if archive_location is None or archive_location == "" or archive_location == 'None':
                logger.warning( "archive_location is empty, setting it to restoreError." % reference)
                archiveHistory = ArchiveHistory(ArchiveStatus.RestoreError,
                                                currentDate,
                                                "",
                                                None)
                self.cdbApi.setArchiveStatus(product.getProductReference(), archiveHistory)
                continue

            #check if location from products table is filled in
            if finalRestoreLocation == None or finalRestoreLocation == "":
                # objUrl ticket subdir will be added by getObj
                logger.warning("Unable to restore product '%s' (reason: no productlocation, setting it to status 'RestoreError'." % reference)
                archiveHistory = ArchiveHistory(ArchiveStatus.RestorError, currentDate, "", None)
                self.cdbApi.setArchiveStatus(product.getProductReference(), archiveHistory)
                continue


            #building  restore groups
            try:
                # Store product for CommVault Restore call.  Products are grouped by subclient first, then by objUrl.
                #TODO; what to do with subtype in subclient
                productsToGet.setdefault(subclient, {}) \
                             .setdefault(finalRestoreLocation, []) \
                             .append((product,  reference, finalRestoreLocation, archive_location))
            except:
                logger.error( "Unable to restore product '%s' from archive (reason: unknown archive ticket %s)." % (reference, archive_location))

                archiveHistory = ArchiveHistory(ArchiveStatus.RestoreError, currentDate, "", None)
                self.cdbApi.setArchiveStatus(reference, archiveHistory)
                # the following line is to purge activeRestoringProducts if could happen
                if reference in self.activeRestoringProducts: self.activeRestoringProducts.remove(reference)


        # Loop over all products to be restored grouped by subclient and objUrl
        # if needed, we have to cache the FTP data first
        for subclient, dct in productsToGet.items():
            for restoreLocation, lst in dct.items():
                
                # Take into account max number of fies to be passed to a Commvault restore
                for i in range(0, len(lst), ltda.ltda_config.RESTORE_MAX_FILES):

                    group = lst[i:i + ltda.ltda_config.RESTORE_MAX_FILES]

                    currentDate = EasyGregorianCalendar(Date())

                    # create a new cache folder
                    if restoreLocation[:6].upper() == 'FTP://':
                        #cache it on the local folder in a unique dir
                        commvault_restore_location = os.path.join(ltda.ltda_config.RESTORE_CACHE_DIR, get_unique_dir())
                    else:
                        #cache it directly on the right disk, in a commvault_cache folder
                        commvault_restore_location = os.path.join(restoreLocation, 'commvault_cache', get_unique_dir())
                    # create archive_location if not exists
                    if not os.path.exists(commvault_restore_location):
                        try:
                            os.makedirs(commvault_restore_location)
                        except:
                            raise Exception('Unable to create restore_cache_location %s ' % commvault_restore_location)


                    # Build list of files to be restored
                    archive_locations_list = []
                    for product, reference, finalRestoreLocation, archive_location in group:
                        archive_locations_list.append(archive_location)

                    try:
                        # Start batch restore operation
                        restoreAction = Restorer(archive_locations_list, commvault_restore_location, ltda.ltda_config.COMMVAULT_CLIENTNAME, ltda.ltda_config.COMMVAULT_BACKUPSETNAME, subclient)
                        commvaultJobNr = restoreAction.do_restore()

                        # Update products and register tickets
                        for product, reference, finalRestoreLocation, archive_location in group:
                            # Update database status to 'Restoring'
                            full_commvault_restore_location = \
                                os.path.join(commvault_restore_location,
                                             os.path.basename(archive_location.rstrip('/').split('/')[-1]))
                            archiveHistory = ArchiveHistory(ArchiveStatus.Restoring, currentDate, commvaultJobNr, full_commvault_restore_location)
                            self.cdbApi.setArchiveStatus(reference, archiveHistory)

                            # Get an updated version of the database product entry
                            product = self.cdbApi.getProductByReference(reference)

                            self.activeRestoringProducts[reference] = ProductTicket(commvaultJobNr, product, self.cdbApi, restoreLocation, commvault_restore_location = full_commvault_restore_location)
                    except:
                        logger.error( "Unable to restore products from archive for subclient %s" % subclient)
                        traceback.print_exc(file=sys.stdout)
                        for product, reference, finalRestoreLocation, archive_location in group:
                            # Update database status to 'RestoreError'
                            archiveHistory = ArchiveHistory(ArchiveStatus.RestoreError, currentDate, "", None)
                            self.cdbApi.setArchiveStatus(reference, archiveHistory)

    # def startCalibrationRestore(self):
    #     logger.debug( "Polling the CDB for calibrations to be restored...")
    #     calibrationData = self.cdbApi.getCalibrationDataByStatus(None,
    #                                                              ArchiveStatus.ToBeRestored,
    #                                                              None)
    #     calibrationsToGet = {}
    #
    #     for calibration in calibrationData:
    #         # JME 18042011 JIRA-156
    #         reference = calibration.getReference()
    #         calibrationLocation = str(calibration.getLocation())
    #         product_type = calibration.getProductType()
    #         currentDate = EasyGregorianCalendar(Date())
    #
    #         if ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS != 0:
    #             logger.debug( "Queue size limit: %d Tickets in queue: %d" % (
    #                 ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS, len(self.activeRestoringProducts)))
    #             if len(self.activeRestoringProducts) >= ltda.ltda_config.RESTORER_MAX_COMMVAULT_REQUESTS:
    #                 break
    #         logger.debug( "Processing calibration data %s." % calibration.getReference())
    #
    #         # get the archive_location from the last archived status from the archivehistory tabel
    #         lastArchive = CalibrationTicket.getLastArchivedStatusFromHistory(calibration, self.cdbApi)
    #         if lastArchive is None:
    #             logger.warning( "Unable to restore calibration data '%s' (reason: never archived), setting it to archive status 'NotArchived' again." % calibration.getReference())
    #             archiveHistory = ArchiveHistory(ArchiveStatus.NotArchived,
    #                                             EasyGregorianCalendar(Date()),
    #                                             "",
    #                                             None)
    #             self.cdbApi.setCalibrationDataArchiveStatus(calibration.getReference(), archiveHistory)
    #             # JME 18042011 no JIRA
    #             continue
    #         archive_location = str(lastArchive.getcommvaultJobNr())
    #
    #         subclient, subdir = getSubclient(product_type)
    #         objUrl = calibration.getLocation()
    #         if objUrl == None or objUrl == "":
    #             # objUrl ticket subdir will be added by getObj
    #             objUrl = ltda.ltda_config.DEFAULT_PROD_REST_FTP
    #
    #         try:
    #             # Store calibration for getObj batch call.  Calibrations are grouped by type first, then by objUrl.
    #             calibrationsToGet.setdefault(subclient, {}) \
    #                              .setdefault(objUrl, []) \
    #                              .append((calibration, reference, calibrationLocation, archive_location))
    #         except:
    #             logger.error( "Unable to restore calibration data '%s' from archive (reason: unknown archive ticket %s)." % (reference, archive_location))
    #
    #             calArchiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.RestoreError, currentDate, "", None)
    #             self.cdbApi.setCalibrationDataArchiveStatus(reference, calArchiveHistory)
    #             # the following line is to purge activeRestoringProducts if could happen
    #             if reference in self.activeRestoringProducts: self.activeRestoringProducts.remove(reference)
    #
    #     # Loop over all calibrations to be restored grouped by type and objUrl
    #     for subclient, dct in calibrationsToGet.items():
    #         for objUrl, lst in dct.items():
    #
    #             # Take into account max number of tickets to be passed to getObj
    #             for i in range(0, len(list), ltda.ltda_config.RESTORE_MAX_FILES):
    #
    #                 group = list[i:i + ltda.ltda_config.RESTORE_MAX_FILES]
    #
    #                 currentDate = EasyGregorianCalendar(Date())
    #
    #                 # Build comma separated list of ticket numbers
    #                 calibration_locations_list = []
    #                 for calibration, reference, calibrationLocation, calibration_location in group:
    #                     calibration_locations_list.append(calibration_location)
    #
    #                 # Configure options to getObj
    #                 options = {
    #                     "subdir" : ltda.ltda_config.GETOBJ_SUBDIR
    #                     }
    #
    #                 try:
    #                     # Start batch restore operation
    #                     restoreAction = Restorer(calibration_locations_list, calibrationLocation,
    #                                              ltda.ltda_config.COMMVAULT_CLIENTNAME,
    #                                              ltda.ltda_config.COMMVAULT_BACKUPSETNAME, subclient)
    #                     commvaultJobNr = restoreAction.do_restore()
    #
    #                     # Update calibrations and register tickets
    #                     for calibration, reference, calibrationLocation, calibration_location in group:
    #                         # Update database status to 'Restoring'
    #                         archiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.Restoring, currentDate, commvaultJobNr, calibrationLocation)
    #                         self.cdbApi.setCalibrationDataArchiveStatus(reference, archiveHistory)
    #
    #                         # Get an updated version of the database calibration entry
    #                         calibration = self.cdbApi.getCalibrationDataByReference(reference)
    #
    #                         # Append directory generated by getObj to store with calibration ticket
    #                         objUrl = objUrl + "/" + ltda.ltda_config.GETOBJ_SUBDIR + "_" + calibration_location
    #
    #                         # Add to list of active tickets
    #                         self.activeRestoringProducts[reference] = CalibrationTicket(commvaultJobNr, calibration, self.cdbApi, objUrl)
    #                 except:
    #                     logger.error( "Unable to restore calibration data from archive (subclient: {%s})." % subclient)
    #                     traceback.print_exc(file=sys.stdout)
    #                     for calibration, reference, calibrationLocation, calibration_location in group:
    #                         # Update database status to 'RestoreError'
    #                         archiveHistory = CalibrationDataArchiveHistory(ArchiveStatus.RestoreError, currentDate, "", None)
    #                         self.cdbApi.setCalibrationDataArchiveStatus(reference, archiveHistory)

    def poll(self):
        while True:
            self.list_active_restores()
            self.startProductRestore()
            self._checkActiveRestores()
            logger.debug("Sleep some time...")
            time.sleep(self.period)


#===============================================================================
# Actual program launch
def main():
    ''' @brief    LTDA Restorer program entry point
    '''


    restorer = LTDA_Restorer()
    restorer.poll()


#===============================================================================
# Global exceptions logging and management

if __name__ == '__main__':
    main()

