from urllib.parse import urlparse
import subprocess
import os
import pathlib
import shutil
import uuid
from time import strftime,localtime

import ltda.ltda_config
import logging

logger = logging.getLogger(__name__)



class CachedData:
    def __init__(self, remote_location, root_cache_dir = ltda.ltda_config.RESTORE_CACHE_DIR):
        self.remote_location = str(remote_location)
        self.parsed_url = urlparse(self.remote_location)
        self.unique_subdir = get_unique_dir()
        self.cache_location= ''

        #set Flags
        self.source_is_FTP = True if self.parsed_url.scheme.upper() == 'FTP' else False
        #self.source_is_file = True if len(pathlib.Path(self.parsed_url.path).suffix) > 0 else False

        #create a new cache_location
        if self.source_is_FTP:
            self.cache_location = os.path.join(root_cache_dir, self.unique_subdir)
        else:
            cache_location = remote_location.rstrip('/')
            #cache_location = '/'.join(remote_location.split('/')[0:-2])  #remove last part of dir/filename
            self.cache_location = os.path.join(cache_location, 'commvault_cache', self.unique_subdir)

        #create archive_location if not exists
        if not os.path.exists(self.cache_location):
            try:
                os.makedirs(self.cache_location)
            except:
                raise Exception('Unable to create archive_cache_location %s ' % self.cache_location)

    def move_to_final_location(self, cache_location):
        #2 cases:
        # 1) if FTP, just upload the data to the final location and clear the cache
        # 2) move the data just one folder up.


        if self.remote_location[0:6].upper() == 'FTP://':
            # upload data, and clear source after copy
            logger.info("uploading cached data %s TO %s" % (cache_location, self.remote_location))
            restore_location = self.upload_data(cache_location)
            self.clear_cache()
        else:
            # move the data just one folder up, source shouldn't be removed because it was moved
            logger.info(
                "moving cached data one folder up %s TO %s" % (cache_location, self.remote_location))
            restore_location = self.move_one_folder_up(cache_location)

        return restore_location



    def upload_data(self, local_file):
        # Download the FTP-data to a local dir before sending it to the archive folder
        self.temp_folder = local_file

        up = Uploader(local_file, self.remote_location)
        if len(pathlib.Path(local_file).suffix) > 0:    #is it file or directory???
            status, filename = up.upload_file()
        else:
            status, filename = up.upload_folder()

        # status is 0 or -1
        if status:
            raise Exception("Unable to upload file/folder from FTP")
        return filename

    def move_one_folder_up(self, source):
        '''move the files/folder from the cache dir one up'''

        fname = os.path.basename(source.rstrip('/'))
        full_destination = os.path.join(self.remote_location, fname)   #comment because not working for a dir

        #clean target if exist
        if os.path.exists(full_destination):
            try:
                if os.path.isfile(full_destination):
                    os.remove(full_destination)
                else:
                    shutil.move(full_destination, full_destination +'_moved_by_archive')
            except:
                raise Exception("Unable to remove %s" % full_destination)

        #move the file/dir
        try:
            shutil.move(source, full_destination)
        except:
            raise Exception("Unable to move file/folder one up : %s" % source)
        try:
            #remove unique dir
            unique_dir = '/'.join(source.rstrip('/').split('/')[0:-1])
            os.rmdir(unique_dir)
        except:
            logger.warning("Unable to delete %s" % unique_dir)
        try:
            #remove commvault_cache dir
            commvault_cache_dir = '/'.join(source.rstrip('/').split('/')[0:-2])
            os.rmdir(commvault_cache_dir)
        except:
            logger.warning("Unable to delete %s" % commvault_cache_dir)
        return full_destination

    def clear_cache(self):
        '''cleanup folder if data is not longer needed'''
        logger.debug('Clearing FTP-cache on folder %s ' % self.temp_folder)
        #check if we have to delete a file or a folder
        if os.path.isdir(self.temp_folder):
            #delete a folder recursively
            try:
                upper_folder = '/'.join(self.temp_folder.rstrip('/').split('/')[0:-1])
                shutil.rmtree(self.temp_folder)
            except:
                logger.warning("Unable to delete temp directory %s " % self.cache_location)
        else:
            try:
                upper_folder = os.path.dirname((self.temp_folder))
                os.remove(self.temp_folder)

            except:
                logger.warning("Unable to delete temp file %s " % self.temp_folder)

        # in case we have removed the last file from the cache dir, we have to delete the upper dir
        try:
            os.rmdir(upper_folder)
        except:
            logger.warning("Unable to delete %s" % upper_folder)

class Uploader:
    def __init__(self, local_path, remote_url):
        self.local_path = str(local_path)
        self.file_name = pathlib.PurePath(self.local_path).parts[-1]
        self.remote_url = str(remote_url)
        self.parsed_remote_url = urlparse(self.remote_url)
        self.FTP_options = 'set net:persist-retries 1;set net:reconnect-interval-base 5; set net:max-retries 2;'

    def upload_folder(self):
        #Upload folder to FTP
        full_remote_path = os.path.join(self.parsed_remote_url.path, self.file_name)
        cmd = 'lftp -e "%s mirror --reverse %s %s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               self.local_path,
               full_remote_path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print (stdout)
            print (stderr)
            return -1, stderr
        #create the full restore path for writing to the db
        restore_dirname = os.path.join(self.remote_url,self.file_name)
        return 0, restore_dirname

    def upload_file(self):
        # Upload to FTP
        f_name = os.path.basename(self.local_path)
        local_dir = os.path.dirname(self.local_path) + '/'  #we should append / to it otherwise it will transfer the dir and not the file in it
        cmd = 'lftp -e "%s;cd %s;lcd %s; put -c %s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               self.parsed_remote_url.path,
               local_dir,
               f_name,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print (stdout)
            print (stderr)
            return -1, stderr

        full_path_restored_file = os.path.join(self.remote_url, os.path.basename(self.local_path))
        return 0, full_path_restored_file

class Downloader:
    def __init__(self, remote_url, local_path):

        self.remote_url = str(remote_url)
        self.parsed_remote_url = urlparse(self.remote_url)
        self.file_name = os.path.basename(self.parsed_remote_url.path.rstrip('/'))
        self.local_path = str(local_path)
        self.full_local_path = os.path.join(self.local_path, self.file_name)

        self.FTP_options = 'set net:persist-retries 1;set dns:fatal-timeout 10;set net:timeout 5;set net:reconnect-interval-base 5; set net:max-retries 2;'

        #set Flags
        self.source_is_FTP = True if self.parsed_remote_url.scheme.upper() == 'FTP' else False
        self.source_is_file = True if len(pathlib.Path(self.parsed_remote_url.path).suffix) > 0 else False

    def clean_and_create_target(self, filepath, isfile):
        #remove if the data already exists, and recreate dir structure
        if os.path.exists(filepath):
            if isfile:
                os.remove(filepath)
            else:
                shutil.rmtree(filepath)

        #create cachedirectory
        dir_2_create = os.path.dirname(filepath)    #cut of last dir or filname
        try:
            if not os.path.exists(dir_2_create):
                os.makedirs(dir_2_create)
        except:
            raise Exception("Unable to create cache_dir %s" % dir_2_create)


    def caching_source(self):

        self.clean_and_create_target(self.full_local_path, self.source_is_file)

        #depending on FTP/local source OR its a file/dir
        if self.source_is_FTP and self.source_is_file:
            status, self.cached_location = self.download_file()
        elif self.source_is_FTP and not self.source_is_file:
            status, self.cached_location = self.download_folder()
        elif not self.source_is_FTP and self.source_is_file:
            self.cached_location = shutil.copy2(self.source_location, self.archive_location)
            status = 0

        elif not self.source_is_FTP and not self.source_is_file:
            self.cached_location = shutil.copytree(self.remote_url, self.full_local_path)
            status = 0
        else:
            raise Exception("invalid state in caching source")
        return status, self.cached_location

    def move_to_final(self, archive_location_dir, fname):
        #fname can be a folder or a file name
        full_path = os.path.join(archive_location_dir, fname)

        #Clean target file/folder if already exists
        isfile = True if len(pathlib.Path(full_path).suffix) > 0 else False
        self.clean_and_create_target(full_path, isfile)

        #reformat archivelocation for dirs and files
        final_archive_location = os.path.join(archive_location_dir,fname) if isfile else archive_location_dir
        status = 0
        logger.debug("moving cache to final : %s -> %s" % (self.full_local_path, final_archive_location ))
        try:
            archive_location = shutil.move(self.full_local_path, final_archive_location)
        except:
            logger.error("Unable to move cache to final location")
            status = -1
        return status, archive_location

    def download_file(self):
        f_name = os.path.basename(self.parsed_remote_url.path)
        # cmd = 'lftp -e "%s lcd %s; mirror --file=%s;exit" -u %s,%s %s' % \
        cmd = 'lftp -e "%s lcd %s; get %s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               self.local_path,
               self.parsed_remote_url.path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print(stdout)
            print(stderr)
            return -1, stderr

        return 0, self.local_path

    def download_folder(self):
        #download from FTP
        cmd = 'lftp -e "lcd %s; mirror %s;exit" -u %s,%s %s' % (
               self.local_path,
               self.parsed_remote_url.path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print(stdout)
            print(stderr)
            return -1, str(stderr)
        if not os.path.exists(self.local_path):
            return -1, stderr
        return 0, ''





class FileTransfer:
    def __init__(self, local_path, remote_url):
        self.local_path = str(local_path)
        self.file_name = pathlib.PurePath(self.local_path).parts[-1]
        self.remote_url = str(remote_url)
        self.parsed_remote_url = urlparse(self.remote_url)
        self.FTP_options = 'set net:reconnect-interval-base 5; set net:max-retries 2;'

    def upload_folder(self):
        #Upload folder to FTP
        full_remote_path = os.path.join(self.parsed_remote_url.path, self.file_name)
        cmd = 'lftp -e "%s mirror --reverse %s %s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               self.local_path,
               full_remote_path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print (stdout)
            print (stderr)
            return -1, stderr
        #create the full restore path for writing to the db
        restore_dirname = os.path.join(self.remote_url,self.file_name)
        return 0, restore_dirname

    def upload_file(self):
        # Upload to FTP
        f_name = os.path.basename(self.local_path)
        local_dir = os.path.dirname(self.local_path) + '/'  #we should append / to it otherwise it will transfer the dir and not the file in it
        cmd = 'lftp -e "%s mirror --reverse --include %s %s %s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               f_name,
               local_dir,
               self.parsed_remote_url.path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print (stdout)
            print (stderr)
            return -1, stderr

        full_path_restored_file = os.path.join(self.remote_url, os.path.basename(self.local_path))
        return 0, full_path_restored_file

    def download_file(self):
        f_name = os.path.basename(self.parsed_remote_url.path)
        cmd = 'lftp -e "%s lcd %s; mirror --file=%s;exit" -u %s,%s %s' % \
              (self.FTP_options,
               os.path.dirname(self.local_path),
               self.parsed_remote_url.path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print(stdout)
            print(stderr)
            return -1, stderr

        return 0, self.local_path

    def download_folder(self):
        #download from FTP
        cmd = 'lftp -e "lcd %s; mirror %s;exit" -u %s,%s %s' % \
              (os.path.dirname(self.local_path),
               self.parsed_remote_url.path,
               self.parsed_remote_url.username,
               self.parsed_remote_url.password,
               self.parsed_remote_url.hostname)
        logger.debug(cmd)
        stdout, stderr, err_nr = run_command2(cmd)
        if err_nr:
            print(stdout)
            print(stderr)
            return -1, str(stderr)
        if not os.path.exists(self.local_path):
            return -1, stderr
        return 0, ''


def get_unique_dir():
        '''it will return a unique folder'''
        return strftime("%Y%m%d", localtime()) + '_%s' % uuid.uuid4()

def run_command2(cmd):
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
    stdout, stderr = p.communicate()
    err_nr = p.returncode
    return str(stdout), str(stderr), err_nr